! $Id: COMBase.f90 2781 2017-02-06 04:33:37Z ian $
! Copyright 2016 M.E.G.M.S.  See LICENCE.txt for licence (Apache 2.0).
! See http://www.megms.com.au/winapi-bindings.htm for more information.
  
!> @file
!! Defines the COMBase module.


!*******************************************************************************
!!
!> Defines the Manager type, that encapsulates operations on an interface 
!! to a COM object.

MODULE COMBase
  
  USE, INTRINSIC :: ISO_C_BINDING, ONLY:  &
      C_PTR,  &
      C_FUNPTR,  &
      C_SHORT,  &
      C_INT,  &
      C_LONG,  &
      C_CHAR,  &
      C_SIGNED_CHAR,  &
      C_NULL_PTR,  &
      C_FLOAT,  &
      C_INT8_T,  &
      C_INT16_T,  &
      C_INT32_T,  &
      C_INT64_T,  &
      C_INTPTR_T
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  !> The integer kind to use for COM api result codes.
  !
  ! Perhaps this should be the value of C_INT32_T.
  INTEGER, PARAMETER, PUBLIC :: HRESULT = C_LONG
  
  !> Wide character integer kind.
  !!
  !! This is used to support UTF-16 strings.
  INTEGER, PARAMETER, PUBLIC :: WCHAR = SELECTED_INT_KIND(4)
  
  !> Floating point real kind.
  INTEGER, PARAMETER, PUBLIC :: FLOAT = C_FLOAT
  
  ! Various integer kinds.
  !
  ! We have made these the bit specific C types, not the general 
  ! integer types that are used in the headers, on the assumption that 
  ! interfaces are binary specifications, regardless of compiler.  The 
  ! trailing comment indicates the general integer type.
  !
  ! Fortran doesn't care about signed versus unsigned, but we differentiate 
  ! to be consistent and for the sake of in-source documentation.  
  INTEGER, PARAMETER, PUBLIC :: BOOL = C_INT32_T        ! int
  INTEGER, PARAMETER, PUBLIC :: UINT8 = C_INT8_T        ! unsigned char
  INTEGER, PARAMETER, PUBLIC :: INT8 = C_INT8_T         ! signed char
  INTEGER, PARAMETER, PUBLIC :: UINT16 = C_INT16_T      ! unsigned short
  INTEGER, PARAMETER, PUBLIC :: INT16 = C_INT16_T       ! short
  INTEGER, PARAMETER, PUBLIC :: UINT32 = C_INT32_T      ! unsigned int
  INTEGER, PARAMETER, PUBLIC :: INT32 = C_INT32_T       ! int
  INTEGER, PARAMETER, PUBLIC :: UINT64 = C_INT64_T      ! __int 64
  ! There are a few cases where things in the headers that are declared 
  ! as UINT32 are then declared as DWORD in the documentation.
  INTEGER, PARAMETER, PUBLIC :: DWORD = C_INT32_T       ! unsigned long
  
  !> Windows handle type.
  INTEGER, PARAMETER, PUBLIC :: HANDLE = C_INTPTR_T
  
  PUBLIC :: UTF8ToUTF16   ! <--- What is this doing in this module?  NFI!
  PUBLIC :: UTF16ToUTF8
  
  !-----------------------------------------------------------------------------
  
  !> A GUID, which are used for many things, including interface identifiers.
  !
  ! From guiddef.h.  Because this is a binary standard, we have used the 
  ! explicitly sized C integer types.  There is probably a particular 
  ! meaning for the various fields, depending on the GUID generator 
  ! in use.
  TYPE, BIND(C), PUBLIC :: GUID
    INTEGER(C_INT32_T) :: x             ! long
    INTEGER(C_INT16_T) :: s1            ! short
    INTEGER(C_SHORT) :: s2              ! short
    INTEGER(C_SIGNED_CHAR) :: c(8)      ! unsigned char
  END TYPE GUID
  
  !-----------------------------------------------------------------------------
  ! "00000000-0000-0000-c000-000000000046" From Unknwn.h
  
  !> The IID for IUnknown.  This is exported by the COM libraries.
  TYPE(GUID), BIND(C, NAME='IID_IUnknown') :: IID_IUnknown
  
  !> The VTable for the IUknown interface.
  TYPE, BIND(C), PUBLIC :: IUnknownVtbl
    !> Query an interface to the object.
    TYPE(C_FUNPTR) :: QueryInterface
    !> Increase the reference count of the object.
    TYPE(C_FUNPTR) :: AddRef
    !> Decrease the reference count of the object.
    TYPE(C_FUNPTR) :: Release
  END TYPE IUnknownVtbl
  
  !-----------------------------------------------------------------------------
  
  !> Fortran object that represents an interface to a COM object.
  !!
  !! The interface of the represented object is IUnknown.  Extensions 
  !! of this type provide for other COM interfaces.
  !!
  !! For a while we called this the Manager type, because it managed a 
  !! COM object, one that happened to be an extension of IUnknown.  
  !! For symmetry with other types it is now named after the base 
  !! interface that it manages, which is IUnknown, and which also 
  !! happens to be the base interface for all other COM interfaces.
  TYPE, PUBLIC :: IUnknown
    !> The C address of the underlying COM object.  We have no idea 
    !! what is at this address (the object could be huuuuge), apart that 
    !! the first component must be the C address of an interface, and the 
    !! first component of that interface must be the C address of the 
    !! VTable for the interface.  See vtable_iunknown for more.
    !
    ! Maybe this is formally the C address of a COM "interface", rather 
    ! than an "object".  Not sure.
    TYPE(C_PTR) :: object = C_NULL_PTR
  CONTAINS
    PROCEDURE, PRIVATE :: manager_assign
    GENERIC :: ASSIGNMENT(=) => manager_assign
    FINAL :: manager_final
    PROCEDURE, NOPASS :: IID => manager_IID
  END TYPE IUnknown
  
  PUBLIC :: Release
  PUBLIC :: Exists
  PUBLIC :: Get
  PUBLIC :: QueryInterface
  PUBLIC :: VTable
  
  PUBLIC :: CoInitialize
  PUBLIC :: CoUninitialize
  PUBLIC :: CoCreateInstance
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Implementation of defined assignment for the manager type.
  !!
  !! The defined assignment ensures that the reference count on the 
  !! underlying object accounts for the thing on the left hand side of 
  !! the assignment.  The dummy argument corresponding to the left hand 
  !! side is INTENT(OUT), which means that it will be finalized prior 
  !! to this procedure being invoked - if the left hand side variable 
  !! previously referenced another object (or the same object) then 
  !! that finalization would have reduced the reference count for that 
  !! object appropriately.
  !!
  !! This procedure is elemental so that it covers the case of 
  !! having variables that hold arrays of managed objects, etc.
  
  IMPURE ELEMENTAL SUBROUTINE manager_assign(lhs, rhs)
    
    USE, INTRINSIC :: ISO_C_BINDING, ONLY:  &
        C_ASSOCIATED,  &
        C_F_POINTER,  &
        C_F_PROCPOINTER
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The object being assigned to.
    CLASS(IUnknown), INTENT(OUT) :: lhs
    
    !> The value being assigned.
    TYPE(IUnknown), INTENT(IN) :: rhs
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: count          ! The reference count of the object.
    
    !***************************************************************************
    
    IF (C_ASSOCIATED(rhs%object)) THEN
      lhs%object = rhs%object
      CALL add_ref_iunknown(lhs%object, count)
    END IF
    
  END SUBROUTINE manager_assign
  
  
  !*****************************************************************************
  !!
  !> Finalizer for Manager objects.
  !!
  !! If an object is currently being managed then it is released.
  !!
  !! Elemental to cover the case of variables that are arrays being 
  !! finalized.
  
  IMPURE ELEMENTAL SUBROUTINE manager_final(obj)
    
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_ASSOCIATED, C_NULL_PTR,  &
        C_F_POINTER, C_F_PROCPOINTER
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The object to finalize.
    TYPE(IUnknown), INTENT(INOUT) :: obj
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: count          ! Release return code.
    
    !***************************************************************************
    
    IF (C_ASSOCIATED(obj%object)) THEN
      CALL release_iunknown(obj%object, count)
      obj%object = C_NULL_PTR
    END IF
    
  END SUBROUTINE manager_final
  
  
  !*****************************************************************************
  !!
  !> Implementation of Manager%%QueryInterface_mgr - queries an interface 
  !! from another object that is compatible with interface handled by the 
  !! extension of manager.
  !!
  !! This default implementation simply queries for the IUnknown interface, 
  !! which should always be available.
  
  FUNCTION manager_IID() RESULT(the_iid)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    !> @returns The interface ID that the manager object deals with.
    !!
    !! This is the IID for IUnknown for this default implementation.
    TYPE(GUID) :: the_iid
    
    !***************************************************************************
    
    the_iid = IID_IUnknown
    
  END FUNCTION manager_IID
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Stand-alone functions for working with Manager objects.
  
  
  !*****************************************************************************
  !!
  !> Calls Release on the underlying COM object.
  !!
  !! Uses the implicit call of the finalizer as part of handling the 
  !! INTENT(OUT) argument.  The C_ASSOCIATED test is simply to avoid a 
  !! variable not used warning.
  
  IMPURE ELEMENTAL SUBROUTINE Release(obj)
    
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_ASSOCIATED
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The object representing the COM object to release.
    CLASS(IUnknown), INTENT(OUT) :: obj
    
    !***************************************************************************
    
    ! Avoid not used warning.
    IF (C_ASSOCIATED(obj%object)) CONTINUE
    
  END SUBROUTINE Release
  
  
  !*****************************************************************************
  !!
  !> Tests whether an underlying object is currently being managed.
  
  IMPURE ELEMENTAL FUNCTION Exists(obj)
    
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_ASSOCIATED
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The object representing the COM object to test.
    CLASS(IUnknown), INTENT(IN) :: obj
    
    !> @returns True if the manager object is currently managing a resource, 
    !! false otherwise.
    LOGICAL :: Exists
    
    !***************************************************************************
    
    Exists = C_ASSOCIATED(obj%object)
    
  END FUNCTION Exists
  
  
  !*****************************************************************************
  !!
  !> Gets the C pointer to the underlying COM object.
  
  IMPURE ELEMENTAL FUNCTION Get(obj)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    !> The object representing the COM object.
    CLASS(IUnknown), INTENT(IN) :: obj
    
    !> @returns The C pointer to the COM object.
    TYPE(C_PTR) :: Get
    
    !***************************************************************************
    
    Get = obj%object
    
  END FUNCTION Get
  
  
  !*****************************************************************************
  !!
  !> Attempts to acquire a different interface for an existing object.
  
  IMPURE ELEMENTAL SUBROUTINE QueryInterface(source, destination)
    
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_ASSOCIATED, C_F_POINTER
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> Object representing the existing COM object.
    CLASS(IUnknown), INTENT(IN) :: source
    
    !> Object that will represent the alternative interface.  If the source 
    !! object does not exist, or the source object does not support the 
    !! interface of the destination type, then the destination object will 
    !! not exist (a call to Exist(destination) will return false).
    CLASS(IUnknown), INTENT(OUT) :: destination
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER(HRESULT) :: hr    ! COM result code - negative on error.
    
    !***************************************************************************
    
    ! No point trying to query an interface from an object that doesn't exist.
    IF (C_ASSOCIATED(source%object)) THEN
      ! If this fails, then destination%object ends up NULL, which is 
      ! can be then tested using Exists.
      CALL query_interface_iunknown(  &
          source%object,  &
          destination%IID(),  &
          destination%object,  &
          hr )
    END IF
    
  END SUBROUTINE QueryInterface
  
  
  !*****************************************************************************
  !!
  !> Gets the C address of the VTable of the interface of an existing object.
  
  FUNCTION VTable(obj)
    
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_ASSOCIATED, C_F_POINTER
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    !> Object representing the existing COM object.
    CLASS(IUnknown), INTENT(IN) :: obj
    
    !> The C address of the VTable.  If the COM object managed by `obj` does 
    !! not exist, this will be C_NULL_PTR.
    TYPE(C_PTR) :: VTable
    
    !***************************************************************************
    
    IF (C_ASSOCIATED(obj%object)) THEN
      VTable = vtable_iunknown(obj%object)
    ELSE
      VTable = C_NULL_PTR
    END IF
    
  END FUNCTION VTable
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for working on IUnknown objects.
  !
  ! These have all been changed (back) to working with the underlying 
  ! object designated by the C address, rather than being passed the 
  ! underlying object as something of type IUnknown.  This is to 
  ! avoid any hypothetical complications from a Fortran compiler doing 
  ! copy-in/copy-out for an argument of type IUnknown - the Fortran compiler 
  ! has no idea what the object looks like so it cannot generate the 
  ! relevant copying code.
  
  
  !*****************************************************************************
  !!
  !> Thin wrapper to the QueryInterface method of IUnknown.
  
  SUBROUTINE query_interface_iunknown(object, riid, ppvObject, hr)
    
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_F_POINTER, C_F_PROCPOINTER
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The C address of the underlying object.
    TYPE(C_PTR), INTENT(IN) :: object
    
    !> The interface identifier of the requested interface.
    TYPE(GUID), INTENT(IN) :: riid
    
    !> The underlying object of the requested interface, if the request 
    !! was successful, otherwise C_NULL_PTR.
    TYPE(C_PTR), INTENT(OUT) :: ppvObject
    
    !> COM status code - negative on error.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ABSTRACT INTERFACE
      FUNCTION QueryInterface_intf(this, riid, ppvObject) RESULT(hr)
        USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_PTR
        IMPORT :: GUID
        IMPORT :: HRESULT
        IMPLICIT NONE
        !-----------------------------------------------------------------------
        TYPE(C_PTR), INTENT(IN) :: this
        TYPE(GUID), INTENT(IN) :: riid
        !DEC$ ATTRIBUTES REFERENCE :: riid
        TYPE(C_PTR), INTENT(OUT) :: ppvObject
        !DEC$ ATTRIBUTES REFERENCE :: ppvObject
        INTEGER(HRESULT) :: hr
        !DEC$ ATTRIBUTES STDCALL :: QueryInterface_intf
      END FUNCTION QueryInterface_intf
    END INTERFACE
    
    ! The VTable of the interface.
    TYPE(IUnknownVtbl), POINTER :: vtbl
    ! Fortran pointer to the QueryInterface method.
    PROCEDURE(QueryInterface_intf), POINTER :: query_proc
    
    !***************************************************************************
    
    CALL C_F_POINTER(vtable_iunknown(object), vtbl)
    CALL C_F_PROCPOINTER(vtbl%QueryInterface, query_proc)
    hr = query_proc(object, riid, ppvObject)
    
  END SUBROUTINE query_interface_iunknown
  
  
  !*****************************************************************************
  !!
  !> Thin wrapper to the AddRef method of IUnknown.
  
  SUBROUTINE add_ref_iunknown(object, count)
    
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_F_POINTER, C_F_PROCPOINTER
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The C address of the underlying object.
    TYPE(C_PTR), INTENT(IN) :: object
    
    !> The reference count after the operation.
    INTEGER(C_LONG), INTENT(OUT) :: count
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ABSTRACT INTERFACE
      FUNCTION AddRef_intf(this) RESULT(count)
        USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_LONG, C_PTR
        IMPLICIT NONE
        !-----------------------------------------------------------------------
        TYPE(C_PTR), INTENT(IN) :: this
        INTEGER(C_LONG) :: count
        !DEC$ ATTRIBUTES STDCALL :: AddRef_intf
      END FUNCTION AddRef_intf
    END INTERFACE
    
    ! The VTable of the interface.
    TYPE(IUnknownVtbl), POINTER :: vtbl
    ! Fortran pointer to the AddRef method.
    PROCEDURE(AddRef_intf), POINTER :: add_proc
    
    !***************************************************************************
    
    CALL C_F_POINTER(vtable_iunknown(object), vtbl)
    CALL C_F_PROCPOINTER(vtbl%AddRef, add_proc)
    count = add_proc(object)
    
  END SUBROUTINE add_ref_iunknown
  
  
  !*****************************************************************************
  !!
  !> Thin wrapper to the Release method of IUnknown.
  
  SUBROUTINE release_iunknown(object, count)
    
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_F_POINTER, C_F_PROCPOINTER
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The C address of the underlying object.
    TYPE(C_PTR), INTENT(IN) :: object
    
    !> The reference count after the operation.
    INTEGER(C_LONG), INTENT(OUT) :: count
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ABSTRACT INTERFACE
      FUNCTION Release_intf(this) RESULT(count)
        USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_LONG, C_PTR
        IMPLICIT NONE
        !-------------------------------------------------------------------------
        TYPE(C_PTR), INTENT(IN) :: this
        INTEGER(C_LONG) :: count
        !DEC$ ATTRIBUTES STDCALL :: Release_intf
      END FUNCTION Release_intf
    END INTERFACE
    
    ! The VTable of the interface.
    TYPE(IUnknownVtbl), POINTER :: vtbl
    ! Fortran pointer to the Release method.
    PROCEDURE(Release_intf), POINTER :: release_proc
    
    !***************************************************************************
    
    CALL C_F_POINTER(vtable_iunknown(object), vtbl)
    CALL C_F_PROCPOINTER(vtbl%Release, release_proc)
    count = release_proc(object)
    
  END SUBROUTINE release_iunknown
  
  
  !*****************************************************************************
  !!
  !> Get the C address of the VTable of the underlying object.
  
  FUNCTION vtable_iunknown(object) RESULT(vtable_ptr)
    
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_F_POINTER
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    !> The C address of the underlying object.
    TYPE(C_PTR), INTENT(IN) :: object
    
    !> The C address of the VTable.
    TYPE(C_PTR) :: vtable_ptr
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Perhaps this should be COMInterface - object/interface confusion.
    TYPE, BIND(C) :: COMObject
      ! The C address of the table of C procedure pointers that implement the 
      ! methods of the interface of the object (the VTable).
      !
      ! In many respects, the VTable for an interface *is* the interface, 
      ! hence the confusion about objects/interfaces above.
      TYPE(C_PTR) :: lpVtbl
    END TYPE COMObject
    
    TYPE(COMObject), POINTER :: object_ptr
    
    !***************************************************************************
    
    CALL C_F_POINTER(object, object_ptr)
    vtable_ptr = object_ptr%lpVtbl
    
  END FUNCTION vtable_iunknown
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Various utility procedures.
  
  
  !*****************************************************************************
  !!
  !> Convert a multibyte UTF-8 string to a wide character (though potentially 
  !! still multibyte) UTF-16 string.
  
  FUNCTION UTF8ToUTF16(narrow_string) RESULT(wide_string)
    
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_INT, C_LONG, C_LOC
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    !> The UTF-8 string to convert.
    CHARACTER(*), INTENT(IN), TARGET :: narrow_string
    
    !> The wide character variant of the string.
    INTEGER(WCHAR), ALLOCATABLE, TARGET :: wide_string(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTERFACE
      FUNCTION MultiByteToWideChar( codepage, dwFlags, lpMultiByteStr,  &
          cbMultiByte, lpWideCharStr, cchWideChar )
        USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_INT, C_PTR
        IMPORT :: DWORD
        IMPLICIT NONE
        !-----------------------------------------------------------------------
        INTEGER(C_INT), INTENT(IN) :: codepage
        INTEGER(DWORD), INTENT(IN) :: dwFlags
        TYPE(C_PTR), INTENT(IN) :: lpMultiByteStr
        INTEGER(C_INT), INTENT(IN) :: cbMultiByte
        TYPE(C_PTR), INTENT(IN) :: lpWideCharStr
        INTEGER(C_INT), INTENT(IN) :: cchWideChar
        INTEGER(C_INT) :: MultiByteToWideChar
        !DEC$ ATTRIBUTES STDCALL :: MultiByteToWideChar
        !DEC$ ATTRIBUTES ALIAS:'MultiByteToWideChar' :: MultiByteToWideChar
        !DEC$ ATTRIBUTES DECORATE :: MultiByteToWideChar
      END FUNCTION MultiByteToWideChar
    END INTERFACE
    
    ! UTF-8 codepage number.
    INTEGER, PARAMETER :: CP_UTF8 = 65001_C_INT
    
    INTEGER(C_INT) :: wide_length
    
    !***************************************************************************
    
    wide_length = MultiByteToWideChar(  &
        CP_UTF8,  &
        0_DWORD,  &
        C_LOC(narrow_string),  &
        LEN(narrow_string),  &
        C_NULL_PTR,  &
        0_C_INT )
    ALLOCATE(wide_string(wide_length))
    wide_length = MultiByteToWideChar(  &
        CP_UTF8,  &
        0_DWORD,  &
        C_LOC(narrow_string),  &
        LEN(narrow_string),  &
        C_LOC(wide_string),  &
        wide_length )
    
  END FUNCTION UTF8ToUTF16
  
  
  !*****************************************************************************
  !!
  !> Convert a multibyte UTF-8 string to a wide character (though potentially 
  !! still multibyte) UTF-16 string.
  
  FUNCTION UTF16ToUTF8(wide_string) RESULT(narrow_string)
    
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_INT, C_LONG, C_LOC
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    !> The wide character variant of the string to convert.
    INTEGER(WCHAR), INTENT(IN), TARGET :: wide_string(:)
    
    !> The UTF-8 string.
    CHARACTER(:), ALLOCATABLE, TARGET :: narrow_string
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTERFACE
      FUNCTION WideCharToMultiByte( codepage, dwFlags, lpWideCharStr,  &
          cchWideChar, lpMultiByteStr, cbMultiByte, lpDefaultChar,  &
          lpUsedDefaultChar )
        USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_INT, C_PTR
        IMPORT :: DWORD
        IMPLICIT NONE
        !-----------------------------------------------------------------------
        INTEGER(C_INT), INTENT(IN) :: codepage
        INTEGER(DWORD), INTENT(IN) :: dwFlags
        TYPE(C_PTR), INTENT(IN) :: lpWideCharStr
        INTEGER(C_INT), INTENT(IN) :: cchWideChar
        TYPE(C_PTR), INTENT(IN) :: lpMultiByteStr
        INTEGER(C_INT), INTENT(IN) :: cbMultiByte
        TYPE(C_PTR), INTENT(IN) :: lpDefaultChar
        TYPE(C_PTR), INTENT(IN) :: lpUsedDefaultChar
        INTEGER(C_INT) :: WideCharToMultiByte
        !DEC$ ATTRIBUTES STDCALL :: WideCharToMultiByte
        !DEC$ ATTRIBUTES ALIAS:'WideCharToMultiByte' :: WideCharToMultiByte
        !DEC$ ATTRIBUTES DECORATE :: WideCharToMultiByte
      END FUNCTION WideCharToMultiByte
    END INTERFACE
    
    ! UTF-8 codepage number.
    INTEGER, PARAMETER :: CP_UTF8 = 65001_C_INT
    
    INTEGER(C_INT) :: narrow_length
    
    !***************************************************************************
    
    narrow_length = WideCharToMultiByte(  &
        CP_UTF8,  &
        0_DWORD,  &
        C_LOC(wide_string),  &
        SIZE(wide_string),  &
        C_NULL_PTR,  &
        0_C_INT,  &
        C_NULL_PTR,  &
        C_NULL_PTR )
    ALLOCATE(CHARACTER(narrow_length) :: narrow_string)
    narrow_length = WideCharToMultiByte(  &
        CP_UTF8,  &
        0_DWORD,  &
        C_LOC(wide_string),  &
        SIZE(wide_string),  &
        C_LOC(narrow_string),  &
        narrow_length,  &
        C_NULL_PTR,  &
        C_NULL_PTR )
    
  END FUNCTION UTF16ToUTF8
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Other stand-alone procedures.
  !
  ! Many of these duplicate function in the Intel provided IFCOM module.
  
  
  !*****************************************************************************
  !!
  !> Initialize COM with a single thread apartment concurrency model.
  
  SUBROUTINE CoInitialize(hr)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> COM result code - S_OK on success, negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTERFACE
      FUNCTION CoInitialize_api(pvReserved)
        USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_PTR
        IMPORT :: HRESULT
        IMPLICIT NONE
        !-----------------------------------------------------------------------
        TYPE(C_PTR), INTENT(IN) :: pvReserved
        INTEGER(HRESULT) :: CoInitialize_api
        !DEC$ ATTRIBUTES STDCALL :: CoInitialize_api
        !DEC$ ATTRIBUTES DECORATE :: CoInitialize_api
        !DEC$ ATTRIBUTES ALIAS:'CoInitialize' :: CoInitialize_api
      END FUNCTION CoInitialize_api
    END INTERFACE
    
    !***************************************************************************
    
    hr = CoInitialize_api(C_NULL_PTR)
    
  END SUBROUTINE CoInitialize
  
  
  !*****************************************************************************
  !!
  !> Uninitialize COM.
  
  SUBROUTINE CoUninitialize(hr)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> COM result code - S_OK on success, negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTERFACE
      FUNCTION CoUninitialize_api()
        USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_PTR
        IMPORT :: HRESULT
        IMPLICIT NONE
        !-----------------------------------------------------------------------
        INTEGER(HRESULT) :: CoUninitialize_api
        !DEC$ ATTRIBUTES STDCALL :: CoUninitialize_api
        !DEC$ ATTRIBUTES DECORATE :: CoUninitialize_api
        !DEC$ ATTRIBUTES ALIAS:'CoUninitialize' :: CoUninitialize_api
      END FUNCTION CoUninitialize_api
    END INTERFACE
    
    !***************************************************************************
    
    hr = CoUninitialize_api()
    
  END SUBROUTINE CoUninitialize
  
  
  !*****************************************************************************
  !!
  !> Create a object associated with a particular class identifier.  This 
  !! is a wrapper around the CoCreateInstance API.
  !!
  !! The object is never part of an aggregate and the execution context is 
  !! always CLSCTX_INPROC_SERVER (code for the object will be in a DLL 
  !! or similar that is executed in the context of the calling process).
  !!
  !! The IID of the initial interface to the object is given by the IID 
  !! binding of the object that will manage the interface pointer.
  
  SUBROUTINE CoCreateInstance(rclsid, ppv, hr)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The class identifier for the object.
    TYPE(GUID), INTENT(IN) :: rclsid
    !> The object that will manage the underlying COM object.
    CLASS(IUnknown), INTENT(OUT) :: ppv
    !> COM result code - S_OK on success, negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTERFACE
      FUNCTION CoCreateInstance_api(rclsid, pUnkOuter, dwClsContext, riid, ppv)
        USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_PTR
        IMPORT :: GUID
        IMPORT :: DWORD
        IMPORT :: HRESULT
        IMPLICIT NONE
        !-----------------------------------------------------------------------
        TYPE(GUID), INTENT(IN) :: rclsid
        !DEC$ ATTRIBUTES REFERENCE :: rclsid
        TYPE(C_PTR), INTENT(IN) :: pUnkOuter
        INTEGER(DWORD), INTENT(IN) :: dwClsContext
        TYPE(GUID), INTENT(IN) :: riid
        !DEC$ ATTRIBUTES REFERENCE :: riid
        TYPE(C_PTR), INTENT(OUT) :: ppv
        !DEC$ ATTRIBUTES REFERENCE :: ppv
        INTEGER(HRESULT) :: CoCreateInstance_api
        !DEC$ ATTRIBUTES STDCALL :: CoCreateInstance_api
        !DEC$ ATTRIBUTES DECORATE :: CoCreateInstance_api
        !DEC$ ATTRIBUTES ALIAS:'CoCreateInstance' :: CoCreateInstance_api
      END FUNCTION CoCreateInstance_api
    END INTERFACE
    
    INTEGER(DWORD), PARAMETER :: CLSCTX_INPROC_SERVER = INT(Z'00000001', DWORD)
    
    !***************************************************************************
    
    hr = CoCreateInstance_api(  &
        rclsid,  &
        C_NULL_PTR,  &              ! Not part of an aggregate.
        CLSCTX_INPROC_SERVER,  &
        ppv%IID(),  &
        ppv%object )
    
  END SUBROUTINE CoCreateInstance
  
END MODULE COMBase
